function [D, Z, iter, hist]= indef_m_scheme_KM(K,M,n_pos,X,Y,tol,opts,msch)

% INDEFINITE m-SCHEME DESIGNED FOR PARTIAL LINEAR RESPONSE EIGENVALUE PROBLEM
%  This function computes n_pos smallest positive eigenvalues of the matrix
%  pair (A,B), where
%
%    A=[ K  0 ], where M and K are n-by-n and positive semidefinite, and one of them
%      [ 0  M ]  is definite,
%    B=[ 0  I_n], where I_n is n-by-n identity matrix
%      [I_n  0 ]
%
%            opts.met=0    opts.met>0
%    n_pos=1    ILOCG       ILOPCG        % single vector    
%    n_pos>1   ILOBCG       ILOBPCG       % block version
%
% Inputs
% ======
%
%       K     array (n-by-n), SPD
%       M     array (n-by-n), SPD
%       n_pos integer, # of desired smallest positive eigenvalues
%       X0    array (n-by-n_pos), initial vectors                              
%       Y0    array (n-by-n_pos), initial vectors
%                             Z0=[X0;Y0] initial approximate eigenvectors
%       tol   is used for the deflation: if maximum of all residuals is  
%             <= tol, then appropriate eigenvalue is conv, and is no
%             longer used in iteration
%       opts  struct, options
%             opts.nitn     integer, # of iterivative steps 
%             opts.met = 0,   (i.e., not preconditioned)
%             opts.met = 1,  preconditioners H^{-1}: 
%                           implemented through K & M's Cholesky decompositions
%             opts.met = 2,  preconditioners approximate H^{-1}: 
%                           implemented through K & M's incomplete Cholesky decompositions
%                           to solve any linear systems K x = b, and M y = c by PCG.
%             opts.met = 3,  preconditioners approximate H^{-1}: 
%                           implemented through solving any linear systems K x = b, and M y = c 
%                           by CG(i.e., without preconditioning).
%             opts.met = 4,  preconditioners approximate H^{-1}: 
%                           implemented through using the diagonals of K & M
%                           to solve any linear systems K x = b, and M y = c by PCG.
%   
%
% Outputs
% =======  
%
%     D      array (n_pos-by-1)     computed smallest positive eigenvalues
%     Z      array (2n-by-n_pos)    computed corresponding eigenvectors    
%     hist   struct               iterative history   
%            hist.res (n_pos-by-iter+1)  hist.res(i,:) records the ith set normalized residuals (in 1-norm)
%                                       ||H*z-D*z||/( ||H||*||z||+|D|*||z|| ) 
%            hist.eigen  (n_pos-by-iter+1)  hist.eigen(i,:) records the ith set of approximate eigenvalues.  
%            hist.cgiter (2-by-1)   defined when 2<=opts.met<=4
%                                hist.cgiter(1)=total number of CG steps in solving K*x=b
%                                hist.cgiter(2)=total number of CG steps in solving M*y=c
%            hist.K_CGITER
%            hist.K_CGITER
%---------------------------------------------------------------------------------------------------------     
%n=size(K,1);
%nev=n_pos;  
nrmK=norm(K,1); nrmM=norm(M,1);
nrmH=max(nrmK,nrmM);
[n, nev] = size(X);
D = zeros(nev, 1);
evs=zeros(1,nev);
errs=zeros(1,nev);
K_CGITER=[];M_CGITER=[];T_CGITER=[];
PHHX=[];PHHY=[];step_pos=[];
%%%%%%%%
 if opts.met==1,
   Rk=chol(K); % Rk'*Rk=K
   Rm=chol(M); % Rm'*Rm=M
elseif opts.met==2,
   Rk=ichol(K,'0'); % Rk'*Rk approx K
   Rm=ichol(M,'0'); % Rm'*Rm approx M
   cgopts.nitn = 20;
   cgopts.tol  = 1.0e-2;
   cgopts.met  = 1;
   cgx0=zeros(n,1);
   hist.cgiter=[0 0];
elseif opts.met==3,
   cgopts.nitn = 50;
   cgopts.tol  = 1.0e-2;
   cgopts.met  = 0;
   cgx0=zeros(n,1);
   cgM=0;
   hist.cgiter=[0 0];
elseif opts.met==4,
   cgopts.nitn = 20;
   cgopts.tol  = 1.0e-2;
   cgopts.met  = 0;
   cgx0=zeros(n,1);
   cgM=0;
   hist.cgiter=[0 0];
elseif opts.met==5,
   R = chol([K, -opts.shift*eye(n); -opts.shift*eye(n), M]);
elseif opts.met==6,
   T = [K, -opts.shift*eye(n); -opts.shift*eye(n), M];
   funT = @(x) [K*x(1:n,:)-opts.shift*x(n+1:end,:); M*x(n+1:end,:)-opts.shift*x(1:n,:)];
   cgopts.nitn = 50;
   cgopts.tol  = 1.0e-1;
   cgopts.met  = 0;
   cgx0=zeros(2*n,1);
   cgM=0;
   hist.cgiter=0;
end
%%%%%%%%%%%%%%%

indX = 1 : nev;
% Initialization
[X, Y,m] = MGS(X, Y);
if (m < nev),  
        error('Initial X does not have at least %d  columns.','nev');
end
Ap = X'*K*X + Y'*M*Y;
Bp = X'*Y + Y'*X;
Ap = (Ap+Ap') / 2; Bp = (Bp+Bp') / 2;
[V0, D0] = eig(Ap, Bp, 'chol');
D=diag(D0);
D = real(D)+imag(D);
[D, ind] = sort(D);
V0 = real(V0(:, ind));
D = D(indX);
V0 = V0(:,indX);
X = X*V0; Y = Y*V0;

       
conv = 0; % later conv counts number of converged eigenvalues (if rel.residual <= tol)
maxiter=opts.nitn;
for iter = 1 : maxiter
    % Deflate conv eigenvalues
    maxr = 0;%
    %conv = 0;
    for j = 1 : conv,
        evs(iter,j) = evs(iter-1,j);
        errs(iter,j) = errs(iter-1,j);
    end
      nrmx=sum(abs(X)); nrmy=sum(abs(Y)); nrmz=nrmx+nrmy;  %%%%%%%
    for j = conv + 1 : nev
        RX(:, j) = K*X(:, j) - D(j)*Y(:, j);
        RY(:, j) = M*Y(:, j) - D(j)*X(:, j);
       r = (norm(RX(:, j),1)+norm(RY(:, j),1))/(nrmH*nrmz(j)+abs(D(j))*nrmz(j));
%         r = norm([RX(:, j); RY(:, j)], 1) / ...
%             (norm([K*X(:, j); M*Y(:,j)],1) + abs(D(j))*norm([X(:, j);
%             Y(:,j)],1));
        ee(j) = D(j);
        rr(j) = r;
        if r > maxr, maxr = r; end%
        if maxr <= tol, 
            conv = conv + 1; 
            fprintf('%d. pos. has converged in %d. step\n',conv,iter)
            step_pos=[step_pos,[conv;iter]];
        end
    end
    if iter>1
         evs=[evs;ee];
         errs=[errs;rr];
    else
        evs=ee;
        errs=rr;
    end
    fprintf('step %3d,  conv %2d,  r = %.2e,  trace = %.14e\n',...
        iter, conv, maxr, sum(D));%
    
    if conv >= nev, break; end
    active = 1 : nev-conv;

    % Generate projection subspace
   % WX = TK*RX(:, conv + active);
   % WY = TM*RY(:, conv + active); 
   if     opts.met==0,
       WX=RX; WY=RY;
   elseif opts.met==1,           
        for  ii=conv+active,                % WX = K\RX, WY=M\RY
            WX(:,ii)=Rk\((Rk')\RX(:,ii));
            WY(:,ii)=Rm\((Rm')\RY(:,ii)); 
        end   
   elseif opts.met==2,
       for  ii=conv+active,
           [WX(:,ii), cg_error, cg_iter, cg_flag] = LinCG(K, cgx0, RX(:,ii), Rk, cgopts);
           K_CGITER=[K_CGITER,cg_iter];
           hist.cgiter (1)=hist.cgiter (1)+cg_iter;
           [WY(:,ii), cg_error, cg_iter, cg_flag] = LinCG(M, cgx0, RY(:, ii), Rm, cgopts);
           M_CGITER=[M_CGITER,cg_iter];
           hist.cgiter (2)=hist.cgiter (2)+cg_iter;
       end 
   elseif opts.met==3 || opts.met==4,
       for ii=conv+active,
           [WX(:,ii), cg_error, cg_iter, cg_flag] = LinCG(K, cgx0, RX(:,ii), cgM, cgopts);
           K_CGITER=[K_CGITER,cg_iter];
           hist.cgiter(1)=hist.cgiter(1)+cg_iter;
           [WY(:,ii), cg_error, cg_iter, cg_flag] = LinCG(M, cgx0, RY(:,ii), cgM, cgopts);
           M_CGITER=[M_CGITER,cg_iter];
           hist.cgiter(2)=hist.cgiter(2)+cg_iter;
       end
   elseif opts.met==5,
       W = R\(R'\[RX(:, conv+active); RY(:, conv+active)]);
       WX = W(1:n, :);
       WY = W(n+1:end, :);
   elseif opts.met==6,
       WX = []; WY = [];
       for ii=conv+active,
           [W(:, 1), cg_error, cg_iter, cg_flag] = LinCG(funT, cgx0, [RX(:,ii); RY(:,ii)], cgM, cgopts);
           T_CGITER=[T_CGITER,cg_iter];
           hist.cgiter(1)=hist.cgiter(1)+cg_iter;
%           [W(:, 2), cg_error, cg_iter, cg_flag] = LinCG(funT, cgx0, [RX(:,ii); -RY(:,ii)], cgM, cgopts);
          
           WX = [WX, W(1:n, :)];
           WY = [WY, W(n+1:end, :)];
       end   
   end
    
      % Search subspace
   if iter == 1 || (isnumeric(msch) && msch==2),
        VX = [X, WX];
        VY = [Y, WY];
   else
     if isnumeric(msch)
       % msch=2,3,4,5,...
%        if msch==2 % PSD/A or BPSD/A
%            PHHX=[];PHHY=[];
%        else % using previous iterates in search subspace
        if msch>2
           PHX = [PX(:, conv + active) PHHX];   
           PHY = [PY(:, conv + active) PHHY];
           if  iter<=(msch-2)
               % not enough previous directions
               PHHX=PHX;PHHY=PHY;
           else
              % e.g. msch=4;iter=3; PH=[P^(3),P^(2),P^(1)]
              % e.g. msch=4;iter=3; PHH=[P^(3),P^(2)]
              sizeact=size(conv+active,2);
              PHHX=PHX(:,1:(msch-2)*sizeact); 
              PHHY=PHY(:,1:(msch-2)*sizeact);
           end %iter<(msch-2)
       end % msch>2   
    else % msch is some string
       % whole history: all previous directions
        sizeact=size(conv+active,2);
        PHHX = [PX(:, conv+active) PHHX];
        PHHY = [PY(:, conv+active) PHHY];
    end % isnumeric(msch)
    %size(PHHX)
    VX=[X, WX, PHHX]; VY = [Y, WY,PHHY]; 
   end
   %size(VX)
% % % %        m = 2*nev-conv;
% % % %        indP = nev + active;
% % %     else
% % %         
% % %             
% % %            
% % % % %         VX = [X, PX(:, conv + active), WX];
% % % % %         VY = [Y, PY(:, conv + active), WY];
% % % %        m = 3*nev-2*conv;
% % % %        indP = [nev + active, 2*nev-conv + active];
% % %     end
    [VX, VY, m] = MGS(VX, VY);
    if m == 0,
        R = rand(n, nev);
        VX = [VX, R];
        VY = [VY, R];
        [VX, VY, m] = MGS(VX, VY);
    end
    indP = nev+1 : m;

    % Solve projected problem
    KK = VX'*K*VX;
    MM = VY'*M*VY;
    Ap = [KK+MM, KK-MM; KK-MM, KK+MM];
    Bp = VX'*VY + VY'*VX;
    Bp = blkdiag(Bp, -Bp); % In fact, Bp = [I, 0; 0, -I]
    Ap = (Ap+Ap') / 2; Bp = (Bp+Bp') / 2;
%    [V0, D0] = eig([KK+MM, KK-MM; -KK+MM, -KK-MM]);
    [V0, D0] = eig(Ap, Bp);
%    [V0, D0] = eig(Bp, Ap, 'chol'); D0 = inv(D0);
    D0 = diag(D0);
    if norm(imag(D0)) > 0,
        plot(diag(D0),'o');
        error(''); 
    end
    D = real(D0)+imag(D0);
    [D, ind] = sort(D);
    V0 = real(V0(:, ind));
    % Extract nev smallest positive eigenvalues
    D = D(m + indX);
    V0 = V0(:, m + indX);
   
    
    %%%%%%%%%%
%     % Infinity norm of every column is 1
%      jj=size(V0,2);
%      [tmp,JJ]=max(abs(V0));
%       for ii=1:jj,
%          V0(:,ii)=V0(:,ii)/V0(JJ(ii),ii);
%       end
    %%%%%%%%%%

    % Update [PX; PY] and [X; Y]
%    PX = X; PY = Y;
%    X = [VX, VX] * V0(:, indX); Y = [VY, -VY] * V0(:, indX);
    PX = VX(:, indP) * (V0(indP, :) + V0(m + indP, :));
    PY = VY(:, indP) * (V0(indP, :) - V0(m + indP, :));
    X = PX + VX(:, indX) * (V0(indX, :) + V0(m + indX, :));
    Y = PY + VY(:, indX) * (V0(indX, :) - V0(m + indX, :));

end

% last residuals and eigenvalues
 RX=K*X -Y*diag(D);
 RY=M*Y - X*diag(D);
 nrmx=sum(abs(X)); nrmy=sum(abs(Y)); nrmz=nrmx+nrmy; 
 r = (sum(abs(RX))+sum(abs(RY)))./(nrmH*nrmz+abs(D').*nrmz); 
 hist.eigen=[evs;D'];
 hist.res=[errs;r];
 hist.K_CGITER=K_CGITER;
 hist.M_CGITER=M_CGITER;
 hist.T_CGITER=T_CGITER;
 hist.step_pos=step_pos;
Z=[Y;X];